library IEEE;
use IEEE.std_logic_1164.all;
use ieee.std_logic_textio.all;
use IEEE.numeric_std.all;
use work.control;

library std;
use std.textio.all;

entity Lab4TestBench is
  
end Lab4TestBench;

architecture Lab4_arch of Lab4TestBench is

  component control is
    generic(
      cntmax : integer			 -- length of cycle
      -- we use this because simulation doesn't require the
      -- millions of cycles that hardware does
      );
    port (
      clk: in STD_LOGIC;                       -- clk input (16 KHz)
      rst: in STD_LOGIC;                       -- reset input
      Addr: out INTEGER range -1 to 65536;     -- output to address
      data: in STD_LOGIC_VECTOR (7 downto 0);  -- input from the data lines of the RAM
      LLed: out STD_LOGIC_VECTOR (7 downto 0); -- ASCII representation for Left LED
      RLed: out STD_LOGIC_VECTOR (7 downto 0); -- ASCII representation Right LED
      ALSel: out std_logic                    -- Address/LED Selection (0=Addr,1=LED)
      );
  end component control;

  component LineSel is
    port (
      LLed: in STD_LOGIC_VECTOR (6 downto 0);  -- intended left LED output
      RLed: in STD_LOGIC_VECTOR (6 downto 0);  -- intended right LED output
      Addr: in integer range -1 to 65536; 	 -- RAM address
      rst: in std_logic;                       -- reset
      clk: in std_logic;                       -- clk
      A: out STD_LOGIC_vector(15 downto 0);    -- output on LED/Address lines
      ALSel: in STD_LOGIC                      -- input for Address/Led Selection (0=Addr,1=LED)
      );
  end component LineSel;
  
  component ascii_led is
    port (
      a: in STD_LOGIC_VECTOR (7 downto 0);
      led: out STD_LOGIC_VECTOR (6 downto 0)
      );
  end component ascii_led;
  
  -----------------------------------------------------
    -- these are component connecting (ie. structural)
    -- signals
    -----------------------------------------------------
    signal char_sig: std_logic_vector (7 downto 0);
  signal LLed_ascii : std_logic_vector (7 downto 0);
  signal RLed_ascii : std_logic_vector (7 downto 0);
  signal LLed_sig : std_logic_vector (6 downto 0);
  signal RLed_sig : std_logic_vector (6 downto 0);
  signal ALSel_sig : std_logic := '0';
  signal reset : std_logic := '0';
  signal Addr_sig : integer range -1 to 65536 := 0;
  signal clk : std_logic := '0';
  signal A_sig: std_logic_vector (15 downto 0);
  
  -----------------------------------------------------
    -- these are signals, constants and shared variables
    -- needed for the testbench, but not generally used
    -- for actual hardware implementation
    -----------------------------------------------------
    constant clk_period : time := 5 ns;
  constant trials : integer := 25;
  constant CNTMAX_const: integer := 7;
  constant ADDR_MIN: integer := 1;

  signal reversible: boolean := true;
  signal done : boolean := false;
  signal done_sim : boolean := false;
  signal file_ready : boolean := false;
  signal cnt : integer := 0;
  shared variable prev_char : std_logic_vector(7 downto 0) := x"00";
  -- preceding character
  shared variable char : std_logic_vector(7 downto 0) := x"00";
  -- the last character inputted as data
  shared variable char_addr: integer;
  -- the last address accessed
  shared variable prev_char_addr: integer;
  -- the previous address accessed
  
  shared variable check : boolean := true;

  type input_array_type is array (natural range <>) of std_logic_vector(7 downto 0);
  shared variable input_data : input_array_type (0 to 30) := (others => X"00");
  -- this is an array that will hold the input sequence
  -- it allows us to access this array using your addressing
  -- output
  
  shared variable failed_for : integer := 0;
  shared variable failed_back: integer := 0;
  shared variable passed_for : integer := 0;
  shared variable passed_back: integer := 0;
  
begin

  control_inst: control
    generic map(cntmax=>CNTMAX_const)
    port map ( 
      clk => clk,
      rst => reset,
      Addr => Addr_sig,
      data => char_sig,
      LLed => LLed_ascii,
      RLed => RLed_ascii,
      ALSel => AlSel_sig
      );   		
  linesel_inst: linesel	
    port map (
      LLed => LLed_sig,
      RLed => RLed_sig,
      Addr => Addr_sig,
      rst => reset,
      clk => clk,
      A => A_sig,
      ALSel => AlSel_sig
      );
  AtoLLed: ascii_led
    port map (
      a => LLed_ascii,
      led => LLed_sig
      );
  AtoRLed: ascii_led
    port map (
      a => RLed_ascii,
      led => RLed_sig
      );

  -----------------------------------------------------------------------------
    reset <= '1',
             '0' after clk_period;

-- the clock upon which we synchronize all of our processes
    clk <= '0' when reset='1' else
           not clk after clk_period when not done
           else '0';
  
  -------------------------------------------------------
           -- DataReadProcess: first we read all the data into an
           --					array that we index by the address
           --					to simulate RAM
           -------------------------------------------------------
           DataReadProcess: process(clk)
             file input: text is "input.txt";
             variable input_line_v: line;
             variable i : integer;
           begin
             if not file_ready then
               i := 0;
               while not endfile(input) loop
                 readline(input, input_line_v);
                 hread(input_line_v, input_data(i));
                 i := i+1;
               end loop;
               input_data(i) := x"ff"; -- mark end of message
               file_ready <= true;
             end if;
           end process;


  --------------------------------------------------------
  -- TestingProcess: the process that will actually test
  --				   the design
  --------------------------------------------------------

  TestingProcess: process(clk)
    variable addr_int_v: integer;
    variable input_char: std_logic_vector(7 downto 0);
  begin
    if file_ready then
      if clk'event and clk='1' then
        addr_int_v := to_integer(unsigned(A_sig));
        -- check that address is valid:
        if addr_int_v < input_data'length 
          and addr_int_v > -1 then
          input_char := input_data(addr_int_v);
        else
          input_char := x"00";
        end if;
        -- send resulting data to the control unit
        char_sig <= input_char;
      end if;
    end if;
  end process;

  ------------------------------------------------------------
  -- CharExpectedProcess: this process will read the char that
  --					should be displayed, using AlSel's
  --					rising edge
  ------------------------------------------------------------

  CharExpectedProcess: process(AlSel_sig)
  begin
    if AlSel_sig'event and AlSel_sig='1' then
      if reset = '0' then
        prev_char := char;
        prev_char_addr := char_addr;
        char := char_sig;
        char_addr := Addr_sig;
        if prev_char = x"31" or char = x"31" 
          or char_addr < ADDR_MIN 
          or prev_char_addr < ADDR_MIN then
          check := false;
        elsif char = x"FF" then -- gone past end of message
          reversible <= false;
        else
          check := true;
        end if;
      end if;
    end if;	
  end process;


  -------------------------------------------------------------
    -- LedCheckProcess: checks the current LED output versus the
    -- 					expected LED output using AlSel's
    --					falling edge
    -- (uncomment the write/writeline commands for debugging
    --  information)
    -------------------------------------------------------------
    LedCheckProcess: process(AlSel_sig)
      variable lout: line;
    begin
      if AlSel_sig'event and AlSel_sig='0' then
        if check then
          if prev_char_addr = char_addr then
            null;
          elsif prev_char_addr > char_addr then
            -- going backwards
            
--				write(lout, string'("Time: "));
--				write(lout, now);
--				writeline(output, lout);
--				
--				write(lout, string'("Left:  "));
--				write(lout, char);
--				write(lout, string'(" actual: "));
--				write(lout, lled_ascii);
--				writeline(output, lout);
--				
--				write(lout, string'("Right: "));
--				write(lout, prev_char);
--				write(lout, string'(" actual: "));
--				write(lout, rled_ascii);
--				writeline(output, lout);
--				
--				write(lout, string'("---------------------------------"));
--				writeline(output, lout);
            if char /= lled_ascii or prev_char /= rled_ascii then
              failed_back := failed_back + 1;
              assert (false)
                report "Outputs do not match in reverse" severity note;
            else
              passed_back := passed_back + 1;
            end if;
          else
            -- going forwards
            
--				write(lout, string'("Time: "));
--				write(lout, now);
--				writeline(output, lout);
--				
--				write(lout, string'("Left:  "));
--				write(lout, prev_char);
--				write(lout, string'(" actual: "));
--				write(lout, lled_ascii);
--				writeline(output, lout);
--				
--				write(lout, string'("Right: "));
--				write(lout, char);
--				write(lout, string'(" actual: "));
--				write(lout, rled_ascii);
--				writeline(output, lout);
--				
--				write(lout, string'("---------------------------------"));
--				writeline(output, lout);

            if char /= rled_ascii or prev_char /= lled_ascii then
              failed_for := failed_for + 1;
              assert (false)
                report "Outputs do not match forwards" severity error;
            else
              passed_for := passed_for + 1;
            end if;
          end if;
        end if;
      end if;
      if ((passed_back + failed_back) = trials)
        or not reversible then
        done_sim <= true;
      end if;
    end process;

  -------------------------------------------------------
      -- ReportingProcess: outputs the results of the testing
      -------------------------------------------------------
      ReportingProcess: process(done_sim)
        variable lout : line;
      begin
        if done_sim then
          write(lout, string'("Note: Forwards output passed "));
          write(lout, passed_for);
          write(lout, string'(" times."));
          writeline(output, lout);
          
          write(lout, string'("Warning: Forwards output failed "));
          write(lout, failed_for);
          write(lout, string'(" times."));
          writeline(output, lout);
          
          write(lout, string'("Note: Backwards output passed "));
          write(lout, passed_back);
          write(lout, string'(" times."));
          writeline(output, lout);
          
          write(lout, string'("Warning: Backwards output failed "));
          write(lout, failed_back);
          write(lout, string'(" times."));
          writeline(output, lout);
          
          write(lout, string'("Note: though it is good to have working" &
                              " backwards output, it is not marked in this testbench"));
          writeline(output, lout);
          
          done <= true;
        end if;
      end process;
  
end architecture;
